#!/usr/bin/env python
# coding: utf-8

# In[172]:


import numpy as np
import cmath
import PIL
from PIL import Image

class Julia:
    r"""
    UTILITY:
    Plots the filled Julia set, K(f), for a given holomorphic rational self-map,
    f, of the extended complex plane. Users can specify the resolution of the
    output image, the depth (i.e number of iterations) per orbit, and the escape
    criterion to display the plot. 
    
    ALGORITHM:
    Given f=p/q with degree d. We find the d+1 fixed points of f (counted with
    multiplicity). Next, for every point p in a discrete grid D=[a,b]x[c,d], we
    compute the closest fixed point to the n-th-iterate of p under f and color it
    accordingly. More availiable details at github.com/kanakdhotre99/filled-julia-set

    INPUT:
    - "p" -- list, numerator polynomial coefficients in order of increasing degree,
    i.e p[0] + p[1]*z**1 + ... + p[n]*z**n

    - "q" -- list, denominator polynomial coefficients in order of increasing degree,
    i.e q[0] + q[1]*z**1 + ... + q[n]*z**n

    - "a" -- int, defines min{real(z): z in D} where D is the discrete grid which
    is to be colored
    
    - "b" -- int, defines max{real(z): z in D} where D is the discrete grid which
    is to be colored
    
    - "c" -- int, defines min{imaginary(z): z in D} where D is the discrete grid which
    is to be colored
    
    - "d" -- int, defines max{imaginary(z): z in D} where D is the discrete grid which
    is to be colored
    
    - "res_x" -- int, (optional - default: 300) defines the width of the plot
    
    - "res_y" -- int, (optional - default: 300) defines the height of the plot

    - "depth" -- int, (optional - default: 50) defines the number of iterations taken
    on a seed before computing it's closest fixed point
     
    - "bail_out" -- int, (optional - default: 1000) defines the allowed upper bound of
    the orbit of a point under the function
      
    - "cols" -- list, (optional - default: [(255, 173, 173),(255, 214, 165),(253, 255, 182),
    (202, 255, 191),(155, 246, 255),(160, 196, 255),(189, 178, 255),(255, 198, 255)])is a
    list of RGB tuples that define the colors in the plot

    OUTPUT:
    An RGB image of the filled Julia set in the complex plane.

    EXAMPLES:
    python: Julia.plot([0.01,0,0,0,0,1],[0,0,0,1])
    >>300x300px 24-bit RGB image with wall time: 16.4s
    
    python: Julia.plot([[complex(-0.090706, 0.27145),0, complex(2.41154,-0.133695),-1]],
    [1],a=-1,b=2,c=-1,d=1,resx=450,resy=300)
    >>450x300px 24-bit RGB image with wall time: 19.4s
    
    python: Julia.plot([complex(-0.744336,0.121198),0,1],[1],resx=500,resy=500,depth=200)
    >>500x500px 24-bit RGB image with wall time 2min 17s
    """
    
    def fixed_points(p,q):
        p,q,r = np.polynomial.Polynomial(p), np.polynomial.Polynomial(q), np.polynomial.Polynomial([0,1])
        s = p - r*q
        return(sorted(list(set(s.roots()))))
    def grid(a,b,c,d,resx,resy):
        xvals = np.linspace(a,b,resx)
        yvals = np.linspace(c,d,resy)
        points = []
        for i in xvals:
            for j in yvals:
                points.append(complex(i,j))
        return(points)
    def orbit(seed,p,q,depth,bail_out):
        def f(p,q,z):
            return( sum([p[i]*z**i for i in range(len(p))])/sum([q[i]*z**i for i in range(len(q))])    )
        o = [seed]
        for i in range(depth):
            try:
                o.append( f(p,q,o[-1]) )
            except Exception:
                pass
        o = [i for i in o if i.real<bail_out and i.imag<bail_out]
        return(o[-1])     
    def closest(x,arr):
        can = arr[0]
        for i in range(len(arr)):
            if abs(x-arr[i])<abs(x-can):
                can = arr[i]     
        if abs(x-can)<1:
            return(can)
        else:
            return(np.inf) 
    def plot(p,q,a=-1,b=1,c=-1,d=1,resx=300,resy=300,depth=50,bail_out=1000, cols=
                     [(255, 173, 173), (255, 214, 165),(253, 255, 182),(202, 255, 191),
                      (155, 246, 255),(160, 196, 255), (189, 178, 255),(255, 198, 255)
                     ]):
        fps = Julia.fixed_points(p,q)
        pallete = cols[0:len(fps)]
        pallete.extend([(53,53,53)])
        if np.polynomial.Polynomial(p).degree()>np.polynomial.Polynomial(q).degree():
            fps.append(np.inf)    
        cgs = []
        points = Julia.grid(a,b,c,d,resx,resy)
        for i in points:
            cgs.append(Julia.closest( Julia.orbit(i,p,q,depth,bail_out),fps) )   
        colors = [pallete[fps.index(cgs[i])] for i in range(len(cgs))]
        colors = np.asarray(colors).reshape(resx,resy,3)
        whiteFrame = 255 * np.ones((resy,resx,3), np.uint8)
        im = Image.fromarray(whiteFrame)
        for i in range(im.width):
            for j in range(im.height):
                im.putpixel( (i,j), tuple(colors[i][j]) )
        return(im)

